from datetime import datetime
from enum import Enum
from typing import Literal, Union

from verse.core import DataModel


class MatchCondition(DataModel):
    """Match condition."""

    exists: bool | None = None
    """Check item exists."""

    if_match: str | None = None
    """If match."""

    if_none_match: str | None = None
    """If none match."""

    if_modified_since: datetime | None = None
    """If modified since."""

    if_unmodified_since: datetime | None = None
    """If unmodified since."""

    if_version_match: str | None = None
    """If version match."""

    if_version_not_match: str | None = None
    """If version not match."""


class BaseIndex(DataModel):
    """Store index."""

    type: str
    """Index type."""

    name: str | None = None
    """Index name."""


class WildcardIndex(BaseIndex):
    """Wildcard index."""

    path: str = "*"
    """Wilcard path."""

    excluded: list[str] = []
    """Excluded paths."""

    type: Literal["wildcard"] = "wildcard"
    """Index type."""


class FieldIndex(BaseIndex):
    """Field index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["field"] = "field"
    """Index type."""


class ExcludeIndex(BaseIndex):
    """Exclude index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["exclude"] = "exclude"


class RangeIndex(BaseIndex):
    """Field index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["range"] = "range"
    """Index type."""


class HashIndex(BaseIndex):
    """Field index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["hash"] = "hash"
    """Index type."""


class AscIndex(BaseIndex):
    """Ascending index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["asc"] = "asc"
    """Index type."""


class DescIndex(BaseIndex):
    """Descending index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["desc"] = "desc"
    """Index type."""


class ArrayIndex(BaseIndex):
    """Array index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["array"] = "array"
    """Index type."""


class TTLIndex(BaseIndex):
    """TTL index."""

    field: str
    """Index field."""

    type: Literal["ttl"] = "ttl"
    """Index type."""


class GeospatialIndex(BaseIndex):
    """Geospatial index."""

    field: str
    """Index field."""

    field_type: str | None = None
    """Field type."""

    type: Literal["geospatial"] = "geospatial"
    """Index type."""


class TextIndex(BaseIndex):
    """Text index."""

    field: str
    """Index field."""

    type: Literal["text"] = "text"
    """Index type."""


class VectorIndexMetric(str, Enum):
    """Vector index metric."""

    DOTPRODUCT = "dotproduct"
    COSINE = "cosine"
    EUCLIDEAN = "euclidean"
    MANHATTAN = "manhattan"
    HAMMING = "hamming"


class VectorIndexStructure(str, Enum):
    """Vector index structure."""

    FLAT = "flat"
    HNSW = "hnsw"
    QUANTIZED_FLAT = "quantized_flat"
    DISKANN = "diskann"


class VectorIndex(BaseIndex):
    """Vector index."""

    field: str | None = None
    """Index field name."""

    dimension: int = 4
    """Vector dimension."""

    metric: VectorIndexMetric = VectorIndexMetric.DOTPRODUCT
    """Index metric."""

    structure: VectorIndexStructure | None = None
    """Vector index structure."""

    nconfig: dict | None = None
    """Native config."""

    type: Literal["vector"] = "vector"
    """Index type."""


class SparseVectorIndex(DataModel):
    """Sparse vector index."""

    field: str | None = None
    """Sparse vector field name."""

    nconfig: dict | None = None
    """Native config."""

    type: Literal["sparse_vector"] = "sparse_vector"
    """Index type."""


class CompositeIndex(BaseIndex):
    """Composite index."""

    fields: list[
        AscIndex
        | DescIndex
        | ArrayIndex
        | GeospatialIndex
        | TextIndex
        | VectorIndex
        | SparseVectorIndex
    ]
    """Composite index fields."""

    type: Literal["composite"] = "composite"
    """Index type."""


class IndexType:
    FIELD = "field"
    COMPOSITE = "composite"
    ASC = "asc"
    DESC = "desc"
    ARRAY = "array"
    EXCLUDE = "exclude"
    TTL = "ttl"
    GEOSPATIAL = "geospatial"
    TEXT = "text"
    VECTOR = "vector"
    SPARSE_VECTOR = "sparse_vector"
    RANGE = "range"
    HASH = "hash"
    WILDCARD = "wildcard"


Index = Union[
    FieldIndex,
    CompositeIndex,
    AscIndex,
    DescIndex,
    ArrayIndex,
    TTLIndex,
    GeospatialIndex,
    TextIndex,
    VectorIndex,
    SparseVectorIndex,
    RangeIndex,
    HashIndex,
    ExcludeIndex,
    WildcardIndex,
    BaseIndex,
]



from verse.ql import Field, Function, Value


class StoreFunctionName:
    EXISTS = "exists"
    NOT_EXISTS = "not_exists"

    IS_DEFINED = "is_defined"
    IS_NOT_DEFINED = "is_not_defined"
    IS_TYPE = "is_type"

    LENGTH = "length"
    CONTAINS = "contains"
    STARTS_WITH = "starts_with"

    ARRAY_LENGTH = "array_length"
    ARRAY_CONTAINS = "array_contains"
    ARRAY_CONTAINS_ANY = "array_contains_any"

    STARTS_WITH_DELIMITED = "starts_with_delimited"

    RANDOM = "random"
    NOW = "now"

    VECTOR_SEARCH = "vector_search"


class StoreFunction:
    @staticmethod
    def exists() -> Function:
        return Function(name=StoreFunctionName.EXISTS)

    @staticmethod
    def not_exists() -> Function:
        return Function(name=StoreFunctionName.NOT_EXISTS)

    @staticmethod
    def is_defined(field: str) -> Function:
        return Function(
            name=StoreFunctionName.IS_DEFINED, args=(Field(path=field),)
        )

    @staticmethod
    def is_not_defined(field: str) -> Function:
        return Function(
            name=StoreFunctionName.IS_NOT_DEFINED, args=(Field(path=field),)
        )

    @staticmethod
    def is_type(field: str, type: str) -> Function:
        return Function(
            name=StoreFunctionName.IS_TYPE,
            args=(
                Field(path=field),
                type,
            ),
        )

    @staticmethod
    def starts_with(field: str, value: str) -> Function:
        return Function(
            name=StoreFunctionName.STARTS_WITH,
            args=(
                Field(path=field),
                value,
            ),
        )

    @staticmethod
    def contains(field: str, value: str) -> Function:
        return Function(
            name=StoreFunctionName.CONTAINS,
            args=(
                Field(path=field),
                value,
            ),
        )

    @staticmethod
    def length(field: str) -> Function:
        return Function(
            name=StoreFunctionName.LENGTH, args=(Field(path=field),)
        )

    @staticmethod
    def array_contains(field: str, value: Value) -> Function:
        return Function(
            name=StoreFunctionName.ARRAY_CONTAINS,
            args=(
                Field(path=field),
                value,
            ),
        )

    @staticmethod
    def array_contains_any(field: str, value: list[Value]) -> Function:
        return Function(
            name=StoreFunctionName.ARRAY_CONTAINS_ANY,
            args=(
                Field(path=field),
                value,
            ),
        )

    @staticmethod
    def array_length(field: str) -> Function:
        return Function(
            name=StoreFunctionName.ARRAY_LENGTH, args=(Field(path=field),)
        )

    @staticmethod
    def starts_with_delimited(
        field: str,
        value: str,
        delimiter: str,
    ) -> Function:
        return Function(
            name=StoreFunctionName.STARTS_WITH_DELIMITED,
            args=(Field(path=field), value, delimiter),
        )

    @staticmethod
    def random() -> Function:
        return Function(name=StoreFunctionName.RANDOM)

    @staticmethod
    def now() -> Function:
        return Function(name=StoreFunctionName.NOW)

    @staticmethod
    def vector_search(
        vector: list[float] | None = None,
        sparse_vector: dict[int, float] | None = None,
        field: str | None = None,
    ):
        return Function(
            name=StoreFunctionName.VECTOR_SEARCH,
            named_args=dict(
                vector=vector, sparse_vector=sparse_vector, field=field
            ),
        )



from verse.core.exceptions import BadRequestError
from verse.internal.storage_core import (
    Attribute,
    SpecialAttribute,
    StoreOperationParser,
)

from ._models import VectorCollectionConfig


def get_collection_config(op_parser: StoreOperationParser):
    config = op_parser.get_config()
    if config is None:
        return None
    if isinstance(config, dict):
        return VectorCollectionConfig.from_dict(config)
    if isinstance(config, VectorCollectionConfig):
        return config
    raise BadRequestError("Collection config format error")


def is_value_field(field: str) -> bool:
    return field == Attribute.VALUE or field == SpecialAttribute.VALUE


def is_metadata_field(field: str) -> bool:
    return field == Attribute.METADATA or field == SpecialAttribute.METADATA



"""
Config Store on ENV file.
"""

__all__ = ["EnvFile"]

from typing import Any

from dotenv import dotenv_values

from verse.core import Context, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import StoreOperation, StoreProvider

from .._constants import DEFAULT_LABEL
from .._helper import normalize_label
from .._models import ConfigItem, ConfigKey


class EnvFile(StoreProvider):
    path: str

    _items: dict
    _init: bool

    def __init__(self, path: str = ".env", **kwargs):
        """Intialize.

        Args:
            path:
                ENV file path, defaults to ".env".
        """
        self.path = path

        self._init = False

    def init(self, context: Context | None = None) -> None:
        if self._init:
            return

        items = dotenv_values(self.path)
        if items is None:
            self._items = dict()
        else:
            self._items = items
        self._init = True

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        # GET value
        if op_parser.op_equals(StoreOperation.GET):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            if label != DEFAULT_LABEL:
                raise BadRequestError(
                    "Custom label in not support in env file"
                )
            if id not in self._items:
                raise NotFoundError
            value = self._items[id]
            result = ConfigItem(
                key=ConfigKey(id=id, label=label),
                value=value,
            )
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            pass
        else:
            return super().run(operation)
        return Response(result=result)



from verse.core.exceptions import BadRequestError
from verse.internal.storage_core import (
    KeyAttribute,
    SpecialAttribute,
    StoreOperationParser,
    StoreFunctionName,
)
from verse.ql import Comparison, ComparisonOp, Field, Function, FunctionNamespace

from ._models import ObjectCollectionConfig, ObjectQueryConfig, ObjectTransferConfig


class QueryArgs:
    prefix: str | None = None
    delimiter: str | None = None
    start_after: str | None = None
    end_before: str | None = None
    limit: int | None = None
    continuation: str | None = None
    paging: bool | None = False
    page_size: int | None = None


def get_collection_config(op_parser: StoreOperationParser):
    config = op_parser.get_config()
    if config is None:
        return None
    if isinstance(config, dict):
        return ObjectCollectionConfig.from_dict(config)
    if isinstance(config, ObjectCollectionConfig):
        return config
    raise BadRequestError("Collection config format error")


def get_query_config(op_parser: StoreOperationParser):
    config = op_parser.get_config()
    if config is None:
        return None
    if isinstance(config, dict):
        return ObjectQueryConfig.from_dict(config)
    if isinstance(config, ObjectQueryConfig):
        return config
    raise BadRequestError("Collection config format error")


def get_transfer_config(
    op_parser: StoreOperationParser,
) -> ObjectTransferConfig | None:
    config = op_parser.get_config()
    if config is None:
        return None
    if isinstance(config, dict):
        return ObjectTransferConfig(**config)
    if isinstance(config, ObjectTransferConfig):
        return config
    raise BadRequestError("Transfer config format error")


def get_query_args(
    op_parser: StoreOperationParser,
) -> QueryArgs:
    args = QueryArgs()
    expr_list = op_parser.get_where_expr_list()
    not_supported = False
    for expr in expr_list:
        if isinstance(expr, Function):
            if expr.namespace == FunctionNamespace.BUILTIN:
                if expr.name == StoreFunctionName.STARTS_WITH:
                    field = expr.args[0].path
                    prefix = expr.args[1]
                    if (
                        field == KeyAttribute.ID
                        or field == SpecialAttribute.ID
                    ):
                        args.prefix = prefix
                    else:
                        not_supported = True
                elif expr.name == StoreFunctionName.STARTS_WITH_DELIMITED:
                    field = expr.args[0].path
                    prefix = expr.args[1]
                    delimiter = expr.args[2]
                    if (
                        field == KeyAttribute.ID
                        or field == SpecialAttribute.ID
                    ):
                        args.prefix = prefix
                        args.delimiter = delimiter
                    else:
                        not_supported = True
                else:
                    not_supported = True
        elif isinstance(expr, Comparison):
            if isinstance(expr.lexpr, Field) and (
                expr.rexpr is None or isinstance(expr.rexpr, str)
            ):
                field = expr.lexpr.path
                value = expr.rexpr
                if field == KeyAttribute.ID or field == SpecialAttribute.ID:
                    if expr.op == ComparisonOp.GT:
                        args.start_after = value
                    elif expr.op == ComparisonOp.LT:
                        args.end_before = value
                    else:
                        not_supported = True
                else:
                    not_supported = True
            elif isinstance(expr.rexpr, Field) and (
                expr.lexpr is None or isinstance(expr.lexpr, str)
            ):
                field = expr.rexpr.path
                value = expr.lexpr
                if field == KeyAttribute.ID or field == SpecialAttribute.ID:
                    if expr.op == ComparisonOp.LT:
                        args.start_after = value
                    elif expr.op == ComparisonOp.GT:
                        args.end_before = value
                    else:
                        not_supported = True
                else:
                    not_supported = True
        else:
            not_supported = True
        if not_supported:
            raise BadRequestError("Condition not supported")

    args.limit = op_parser.get_limit()
    args.continuation = op_parser.get_continuation()
    config = get_query_config(op_parser)
    if config is not None:
        args.paging = config.paging
        args.page_size = config.page_size
    return args



"""
Local Secret Store on SQLite.
"""

__all__ = ["Local"]


from verse.core.exceptions import ConflictError
from verse.storage.document_store import DocumentStore
from verse.storage.document_store.providers.sqlite import SQLite

from .document_store_provider import DocumentStoreProvider


class Local(DocumentStoreProvider):
    database: str
    table: str

    def __init__(
        self,
        database: str = ":memory:",
        table: str = "secret",
        **kwargs,
    ):
        """Initialize.

        Args:
            database:
                SQLite database, defaults to ":memory:".
            table:
                SQLite table name, defaults to "secret".
        """
        self.database = database
        self.table = table
        store = DocumentStore(
            provider=SQLite(database=self.database, table=self.table)
        )
        try:
            store.create_collection()
        except ConflictError:
            pass
        super().__init__(store=store)



from verse.core.exceptions import BadRequestError

from ._operation import StoreOperation
from ._operation_parser import StoreOperationParser


class Validator:
    @staticmethod
    def validate_batch(
        op_parsers: list[StoreOperationParser],
        allowed_ops: list[str],
        single_collection: bool | None = False,
        single_pk: bool | None = False,
    ):
        if len(op_parsers) == 0:
            raise BadRequestError("BATCH must have at least one operation")
        collections = []
        for op_parser in op_parsers:
            collection = op_parser.get_collection_name()
            if collection not in collections:
                collections.append(collection)
            op = op_parser.get_name()
            if op not in allowed_ops:
                raise BadRequestError(f"{op} not supported in BATCH")
            if (
                op == StoreOperation.PUT
                or op == StoreOperation.DELETE
                or StoreOperation.UPDATE
            ):
                where = op_parser.get_where()
                if where is not None:
                    raise BadRequestError(
                        f"BATCH does not support where clause in {op}"
                    )
            else:
                raise BadRequestError(f"{op} not supported in BATCH")

        if single_collection:
            if len(collections) > 1:
                raise BadRequestError(
                    "BATCH can operate on only a single collection"
                )

    @staticmethod
    def validate_transact(
        op_parsers: list[StoreOperationParser],
        allowed_ops: list[str],
        single_collection: bool | None = False,
        single_pk: bool | None = False,
    ):
        if len(op_parsers) == 0:
            raise BadRequestError("TRANSACT must have at least one operation")
        collections = []
        for op_parser in op_parsers:
            collection = op_parser.get_collection_name()
            if collection not in collections:
                collections.append(collection)
            op = op_parser.get_name()
            if op not in allowed_ops:
                raise BadRequestError(f"{op} not supported in TRANSACT")

        if single_collection:
            if len(collections) > 1:
                raise BadRequestError(
                    "TRANSACT can operate on only a single collection"
                )



"""
Document Store on Amazon DynamoDB.
"""

from __future__ import annotations

__all__ = ["AmazonDynamoDB"]

import json
from decimal import Decimal
from typing import Any

import boto3
from botocore.exceptions import ClientError

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    ParameterParser,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    OrderByDirection,
    Select,
    Update,
    UpdateOp,
    Value,
)

from .._helper import (
    build_item_from_parts,
    build_item_from_value,
    build_query_result,
    get_collection_config,
)
from .._models import DocumentCollectionConfig, DocumentFieldType, DocumentItem


class AmazonDynamoDB(StoreProvider):
    collection: str | None
    table: str | None
    region_name: str | None
    profile_name: str | None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    id_map_field: str | dict | None
    pk_map_field: str | dict | None
    etag_embed_field: str | dict | None
    suppress_fields: list[str] | None
    nparams: dict[str, Any]

    _resource: Any
    _collection_cache: dict[str, DynamoDBCollection]

    def __init__(
        self,
        table: str | None = None,
        region_name: str | None = None,
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        id_map_field: str | dict | None = "id",
        pk_map_field: str | dict | None = "pk",
        etag_embed_field: str | dict | None = "_etag",
        suppress_fields: list[str] | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            table:
                DynamoDB table name mapped to document store collection.
            region_name:
                AWS region name.
            profile_name:
                AWS profile name.
            aws_access_key_id:
                AWS access key id.
            aws_secret_access_key:
                AWS secret access key.
            aws_session_token:
                AWS session token.
            id_map_field:
                Field in the document to map into id.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            pk_map_field:
                Field in the document to map into pk.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            etag_embed_field:
                Field to store the generated ETAG value.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            suppress_fields:
                List of fields to supress when results are returned.
            nparams:
                Native parameters to boto3 client.
        """
        self.collection = table
        self.table = table
        self.region_name = region_name
        self.profile_name = profile_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field
        self.etag_embed_field = etag_embed_field
        self.suppress_fields = suppress_fields
        self.nparams = nparams

        self._resource = None
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._resource is not None:
            return

        resource = None
        if self.profile_name is not None:
            resource = boto3.resource(
                "dynamodb",
                region_name=self.region_name,
                profile_name=self.profile_name,
                **self.nparams,
            )
        elif (
            self.aws_access_key_id is not None
            and self.aws_secret_access_key is not None
        ):
            resource = boto3.resource(
                "dynamodb",
                region_name=self.region_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                **self.nparams,
            )
        else:
            resource = boto3.resource(
                "dynamodb",
                region_name=self.region_name,
                **self.nparams,
            )

        self._resource = resource
        self._conditional_check_failed_exception = (
            resource.meta.client.exceptions.ConditionalCheckFailedException
        )

    def _get_table_name(self, op_parser: StoreOperationParser) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        table = (
            collection_name.split(".")[0]
            if collection_name is not None
            else self.collection
        )
        return table

    def _get_collections(
        self, op_parser: StoreOperationParser
    ) -> list[DynamoDBCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[DynamoDBCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(self._get_collections(single_op_parser))
            return collections
        table = self._get_table_name(op_parser)
        if table is None:
            raise BadRequestError("Collection name must be specified")
        if table in self._collection_cache:
            return [self._collection_cache[table]]
        client = self._resource.Table(table)
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field, table
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field, table
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, table
        )
        col = DynamoDBCollection(
            client,
            ClientHelper,
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._collection_cache[table] = col
        return [col]

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[StoreOperation.PUT, StoreOperation.DELETE],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collections = self._get_collections(op_parser)
        ncall, state = self._get_ncall(
            op_parser, collections, ResourceHelper(self._resource)
        )
        if ncall is None:
            return super().run(operation=operation, context=context)
        nresult, nerror = ncall.invoke(return_error=True)
        result = self._convert_nresult(
            nresult, nerror, state, op_parser, collections
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collections: list[DynamoDBCollection],
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if len(collections) == 1:
            converter = collections[0].converter
            client = collections[0].client
            helper = collections[0].helper
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args = {
                "table": self._get_table_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "table": self._get_table_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_key())
            call = NCall(client.get_item, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, state = converter.convert_put(
                op_parser.get_key(),
                op_parser.get_value(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
            )
            call = NCall(client.put_item, args, nargs)
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            args, state = converter.convert_update(
                op_parser.get_key(),
                op_parser.get_set(),
                op_parser.get_where(),
                op_parser.get_returning(),
            )
            call = NCall(client.update_item, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = converter.convert_delete(
                op_parser.get_key(), op_parser.get_where()
            )
            call = NCall(client.delete_item, args, nargs)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
                collection=op_parser.get_collection_name(),
            )
            args["nargs"] = nargs
            call = NCall(helper.query, args, None)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
                collection=op_parser.get_collection_name(),
            )
            args["nargs"] = nargs
            call = NCall(helper.count, args, None)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, state = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            call = NCall(helper.batch, args, None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            args, state = OperationConverter.convert_transact(
                op_parser.get_operation_parsers(),
                [col.converter for col in collections],
            )
            call = NCall(resource_helper.transact, args, None)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        nerror: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collections: list[DynamoDBCollection],
    ) -> Any:
        if len(collections) == 1:
            processor = collections[0].processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            if nerror is not None:
                if (
                    isinstance(nerror, ClientError)
                    and nerror.response["Error"]["Code"]
                    == "ResourceInUseException"
                ):
                    pass
                else:
                    raise nerror
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            if nerror is not None:
                raise nerror
            result = None
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            if nerror is not None:
                raise nerror
            result = [t.name for t in nresult]
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            if nerror is not None:
                raise nerror
            if "Item" not in nresult:
                raise NotFoundError
            document = self._convert_to_json_dict(nresult["Item"])
            result = build_item_from_value(
                processor=processor,
                value=document,
                include_value=True,
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            if nerror is not None:
                if (
                    isinstance(nerror, ClientError)
                    and nerror.response["Error"]["Code"]
                    == "ConditionalCheckFailedException"
                ):
                    if "Item" in nerror.response:
                        raise PreconditionFailedError
                    raise PreconditionFailedError
                raise nerror
            result = build_item_from_value(
                processor=processor,
                value=state["value"] if state is not None else None,
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            where = op_parser.get_where()
            if nerror is not None:
                if (
                    isinstance(nerror, ClientError)
                    and nerror.response["Error"]["Code"]
                    == "ConditionalCheckFailedException"
                ):
                    if "Item" in nerror.response:
                        raise PreconditionFailedError
                    if where is None:
                        raise NotFoundError
                    else:
                        raise PreconditionFailedError
                elif (
                    isinstance(nerror, ClientError)
                    and nerror.response["Error"]["Code"]
                    == "ValidationException"
                    and "document path" in nerror.response["Error"]["Message"]
                ):
                    if where is None:
                        raise NotFoundError
                    else:
                        raise PreconditionFailedError
                raise nerror
            if "Attributes" in nresult:
                document = self._convert_to_json_dict(nresult["Attributes"])
                result = build_item_from_value(
                    processor=processor, value=document, include_value=True
                )
            else:
                key = op_parser.get_key()
                result = build_item_from_parts(
                    processor=processor,
                    key=key,
                    etag=(
                        state["etag"]
                        if state is not None and "etag" in state
                        else None
                    ),
                )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            where = op_parser.get_where()
            if nerror is not None:
                if (
                    isinstance(nerror, ClientError)
                    and nerror.response["Error"]["Code"]
                    == "ConditionalCheckFailedException"
                ):
                    if "Item" in nerror.response:
                        raise PreconditionFailedError
                    if where is None:
                        raise NotFoundError
                    else:
                        raise PreconditionFailedError
                raise nerror
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            if nerror is not None:
                raise nerror
            result = build_query_result(
                self._parse_query_response(
                    nresult,
                    processor,
                    op_parser.get_limit(),
                    op_parser.get_offset(),
                )
            )
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            if nerror is not None:
                raise nerror
            result = self._parse_count_response(nresult)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            if nerror is not None:
                raise nerror
            result = []
            if state is not None:
                for st in state["values"]:
                    if st is not None:
                        result.append(
                            build_item_from_value(
                                processor=processor,
                                value=st["value"],
                            )
                        )
                    else:
                        result.append(None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            if nerror is not None:
                if (
                    isinstance(nerror, ClientError)
                    and nerror.response["Error"]["Code"]
                    == "TransactionCanceledException"
                ):
                    raise ConflictError
                raise nerror
            result = []
            op_parsers = op_parser.get_operation_parsers()
            for i in range(0, len(op_parsers)):
                if state is not None:
                    st = state["values"][i]
                result.append(
                    self._convert_nresult(
                        {}, None, st, op_parsers[i], [collections[i]]
                    )
                )
        return result

    def _parse_query_response(
        self,
        responses: list,
        processor: ItemProcessor,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[DocumentItem]:
        documents = []
        for response in responses:
            if "Items" in response:
                for item in response["Items"]:
                    document = self._convert_to_json_dict(item)
                    documents.append(
                        build_item_from_value(
                            processor=processor,
                            value=document,
                            include_value=True,
                        )
                    )

        end = len(documents)
        start = 0
        if limit is not None:
            if offset is not None:
                start = offset
                end = start + limit
            else:
                end = limit
        return documents[start:end]

    def _parse_count_response(self, responses: list) -> int:
        count = 0
        for response in responses:
            if "Count" in response:
                count = response["Count"]
        return count

    def _convert_to_json_dict(self, document: dict) -> dict:
        class DecimalEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, Decimal):
                    return int(obj) if obj % 1 == 0 else float(obj)
                return json.JSONEncoder.default(self, obj)

        return json.loads(json.dumps(document, cls=DecimalEncoder))


class ResourceHelper:
    resource: Any

    def __init__(self, resource: Any):
        self.resource = resource

    def create_collection(
        self,
        table: str,
        config: DocumentCollectionConfig | None,
        nargs: Any,
    ):
        def get_attribute_type(type: str) -> str:
            if type == DocumentFieldType.STRING:
                return "S"
            elif type == DocumentFieldType.NUMBER:
                return "N"
            raise BadRequestError("Field type not supported")

        attribute_definitions = []
        key_schema = []
        pk_field = (
            config.pk_field
            if config is not None and config.pk_field is not None
            else "pk"
        )
        pk_type = (
            config.pk_type
            if config is not None and config.pk_type is not None
            else DocumentFieldType.STRING
        )
        attribute_definitions.append(
            {
                "AttributeName": pk_field,
                "AttributeType": get_attribute_type(pk_type),
            }
        )
        key_schema.append({"AttributeName": pk_field, "KeyType": "HASH"})
        id_field = (
            config.id_field
            if config is not None and config.id_field is not None
            else "id"
        )
        id_type = (
            config.id_type
            if config is not None and config.id_type is not None
            else DocumentFieldType.STRING
        )
        attribute_definitions.append(
            {
                "AttributeName": id_field,
                "AttributeType": get_attribute_type(id_type),
            }
        )
        key_schema.append({"AttributeName": id_field, "KeyType": "RANGE"})
        args = {
            "TableName": table,
            "AttributeDefinitions": attribute_definitions,
            "KeySchema": key_schema,
            "BillingMode": "PAY_PER_REQUEST",
        }
        db_table = NCall(
            self.resource.create_table,
            args,
            nargs,
        ).invoke()
        NCall(db_table.wait_until_exists, None, nargs).invoke()

    def drop_collection(self, table: str, nargs: Any):
        NCall(
            self.resource.meta.client.delete_table, {"TableName": table}, nargs
        ).invoke()
        waiter = NCall(
            self.resource.meta.client.get_waiter, ["table_not_exists"], nargs
        ).invoke()
        NCall(waiter.wait, {"TableName": table}, nargs).invoke()

    def list_collections(self, nargs) -> Any:
        response = NCall(self.resource.tables.all, None, nargs).invoke()
        return response

    def transact(self, ops: list[dict]) -> Any:
        nresult = self.resource.meta.client.transact_write_items(
            TransactItems=ops
        )
        return nresult

    def close(self, nargs: Any) -> Any:
        pass


class ClientHelper:
    client: Any

    def __init__(self, client: Any) -> None:
        self.client = client

    def query(self, query: Query, nargs: Any) -> Any:
        call = self._get_query_plan(query, nargs)
        return self._execute_query(call)

    def count(self, query: Query, nargs: Any) -> Any:
        call = self._get_query_plan(query, nargs)
        return self._execute_query(call)

    def batch(self, ops: list[dict]) -> Any:
        with self.client.batch_writer() as batch:
            for op in ops:
                if op["id"] == StoreOperation.PUT:
                    batch.put_item(Item=op["args"]["Item"])
                elif op["id"] == StoreOperation.DELETE:
                    batch.delete_item(Key=op["args"]["Key"])

    def _get_query_plan(self, query: Query, nargs: Any) -> NCall:
        args: dict[str, Any] = {}
        if query.index_name is not None and "." in query.index_name:
            splits = query.index_name.split(".")
            args["IndexName"] = splits[1]
            args["TableName"] = splits[0]
        if query.select is not None:
            args["Select"] = query.select
        if query.projection_expression is not None:
            args["ProjectionExpression"] = query.projection_expression
        if query.key_condition_expression is not None:
            args["KeyConditionExpression"] = query.key_condition_expression
        if query.filter_expression is not None:
            args["FilterExpression"] = query.filter_expression
        if (
            query.expression_attribute_names is not None
            and len(query.expression_attribute_names) > 0
        ):
            args["ExpressionAttributeNames"] = query.expression_attribute_names
        if (
            query.expression_attribute_values is not None
            and len(query.expression_attribute_values) > 0
        ):
            args["ExpressionAttributeValues"] = (
                query.expression_attribute_values
            )
        if query.scan_index_forward is not None:
            args["ScanIndexForward"] = query.scan_index_forward

        limit: int | None = None
        if query.limit is not None:
            if query.offset is not None:
                limit = query.limit + query.offset
            else:
                limit = query.limit
        if limit is not None:
            args["Limit"] = limit

        if query.action == IndexManager.ACTION_QUERY:
            return NCall(self.client.query, args, nargs)
        return NCall(self.client.scan, args, nargs)

    def _execute_query(self, call: NCall) -> Any:
        responses = []
        last_evaluated_key = None
        count = 0
        limit = call.get_arg("Limit")
        while True:
            if last_evaluated_key is not None:
                call.set_arg("ExclusiveStartKey", last_evaluated_key)
            response: dict | None = None
            response = call.invoke()
            responses.append(response)
            if response is not None and "LastEvaluatedKey" in response:
                last_evaluated_key = response["LastEvaluatedKey"]
            else:
                last_evaluated_key = None
            if response is not None and "Count" in response:
                count = count + response["Count"]
            if last_evaluated_key is None:
                break
            if limit is not None and count >= limit:
                break
        return responses


class OperationConverter:
    processor: ItemProcessor
    index_manager: IndexManager

    def __init__(self, processor: ItemProcessor, index_manager: IndexManager):
        self.processor = processor
        self.index_manager = index_manager

    @staticmethod
    def convert_transact(
        op_parsers: list[StoreOperationParser],
        converters: list[OperationConverter],
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for i in range(0, len(op_parsers)):
            op_parser = op_parsers[i]
            converter = converters[i]
            if op_parser.op_equals(StoreOperation.PUT):
                args, state = converter.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append({"Put": args})
                states.append(state)
            elif op_parser.op_equals(StoreOperation.UPDATE):
                args, state = converter.convert_update(
                    op_parser.get_key(),
                    op_parser.get_set(),
                    op_parser.get_where(),
                    op_parser.get_returning(),
                )
                if "ReturnValues" in args:
                    # ReturnValues not supported in update
                    args.pop("ReturnValues")
                ops.append({"Update": args})
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = converter.convert_delete(
                    op_parser.get_key(), op_parser.get_where()
                )
                ops.append({"Delete": args})
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.PUT):
                args, state = self.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append({"id": StoreOperation.PUT, "args": args})
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = self.convert_delete(
                    op_parser.get_key(), op_parser.get_where()
                )
                ops.append({"id": StoreOperation.DELETE, "args": args})
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_get(self, key: Value) -> dict:
        dbkey = self.processor.get_key_from_key(key=key)
        return {"Key": dbkey}

    def convert_put(
        self,
        key: Value,
        value: dict,
        where: Expression | None,
        exists: bool | None,
    ) -> tuple[dict, dict | None]:
        document = self.processor.add_embed_fields(value, key)
        document = json.loads(json.dumps(document), parse_float=Decimal)
        args: dict = {
            "TableName": self.index_manager.table_name,
            "Item": document,
        }
        if where is None:
            pass
        elif exists is False:
            condition_expression = (
                f"attribute_not_exists({self.processor.id_embed_field})"
            )
            args = args | {"ConditionExpression": condition_expression}
        elif exists is True:
            condition_expression = (
                f"attribute_exists({self.processor.id_embed_field})"
            )
            args = args | {"ConditionExpression": condition_expression}
        elif where is not None:
            attribute_names: dict[str, str] = {}
            attribute_values: dict[str, Any] = {}
            condition_expression = f"""
                attribute_exists({self.processor.id_embed_field}) and {
                self.convert_expr(
                where, attribute_names, attribute_values)}"""
            args = args | {
                "ConditionExpression": condition_expression,
                "ReturnValuesOnConditionCheckFailure": "ALL_OLD",
            }
            if len(attribute_names) > 0:
                args["ExpressionAttributeNames"] = attribute_names
            if len(attribute_values) > 0:
                args["ExpressionAttributeValues"] = attribute_values
        return args, {"value": document}

    def convert_update(
        self,
        key: Value,
        set: Update,
        where: Expression | None,
        returning: bool | None,
    ) -> tuple[dict, dict | None]:
        state = None
        dbkey = self.processor.get_key_from_key(key=key)
        if self.processor.needs_local_etag():
            etag = self.processor.generate_etag()
            set = self.processor.add_etag_update(set, etag)
            state = {"etag": etag}
        attribute_names: dict = {}
        attribute_values: dict = {}
        update_expression = self.convert_update_ops(
            set, attribute_names, attribute_values
        )
        args = {
            "TableName": self.index_manager.table_name,
            "Key": dbkey,
            "UpdateExpression": update_expression,
            "ReturnValuesOnConditionCheckFailure": "ALL_OLD",
        }
        if returning is True:
            args = args | {"ReturnValues": "ALL_NEW"}
        elif returning is False:
            args = args | {"ReturnValues": "ALL_OLD"}
        if where is not None:
            condition_expression = self.convert_expr(
                where, attribute_names, attribute_values
            )
        else:
            condition_expression = (
                f"attribute_exists({self.processor.id_embed_field})"
            )
        args["ConditionExpression"] = condition_expression
        if len(attribute_names) > 0:
            args["ExpressionAttributeNames"] = attribute_names
        if len(attribute_values) > 0:
            args["ExpressionAttributeValues"] = attribute_values
        return args, state

    def convert_delete(self, key: Value, where: Expression | None) -> dict:
        dbkey = self.processor.get_key_from_key(key=key)
        args = {
            "TableName": self.index_manager.table_name,
            "Key": dbkey,
            "ReturnValuesOnConditionCheckFailure": "ALL_OLD",
        }
        if where is not None:
            attribute_names: dict = {}
            attribute_values: dict = {}
            condition_expression = self.convert_expr(
                where, attribute_names, attribute_values
            )
            args["ConditionExpression"] = condition_expression
            if len(attribute_names) > 0:
                args["ExpressionAttributeNames"] = attribute_names
            if len(attribute_values) > 0:
                args["ExpressionAttributeValues"] = attribute_values
        else:
            args["ConditionExpression"] = (
                f"""attribute_exists({self.processor.id_embed_field})"""
            )
        return args

    def convert_query(
        self,
        select: Select | None = None,
        where: Expression | None = None,
        order_by: OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
    ) -> dict:
        explicit_index = self._get_explicit_index(collection)
        order_by_field = self._get_order_by_field(order_by)
        select_fields = self._get_select_fields(select)
        where_key_fields, where_non_key_fields = self._get_where_fields(where)
        query_plan = self.index_manager.get_query_plan(
            order_by_field,
            select_fields,
            where_key_fields,
            where_non_key_fields,
            explicit_index,
        )
        query = Query()
        query.action = query_plan.action
        query.index_name = explicit_index
        query.expression_attribute_names = {}
        query.expression_attribute_values = {}
        query.limit, query.offset = self.convert_limit_offset(limit, offset)
        query.select, query.projection_expression = self.convert_select(
            select, query.expression_attribute_names
        )
        if query_plan.action == IndexManager.ACTION_QUERY:
            query.index_name = query_plan.index.name
            if query.index_name == "$main":
                query.index_name = None
            ke, fe = self._split_key_filter_expression(where, query_plan.index)
            query.key_condition_expression = self.convert_expr(
                ke,
                query.expression_attribute_names,
                query.expression_attribute_values,
            )
            query.filter_expression = self.convert_expr(
                fe,
                query.expression_attribute_names,
                query.expression_attribute_values,
            )
            query.scan_index_forward = self.convert_order_by(order_by)
        elif query_plan.action == IndexManager.ACTION_SCAN:
            if where is not None:
                query.filter_expression = self.convert_expr(
                    where,
                    query.expression_attribute_names,
                    query.expression_attribute_values,
                )
        return {"query": query}

    def convert_count(
        self, where: Expression | None, collection: str | None
    ) -> dict:
        args = self.convert_query(where=where, collection=collection)
        args["query"].select = "COUNT"
        return args

    def convert_update_ops(
        self, update: Update, attribute_names, attribute_values
    ) -> str:
        sets: list[str] = []
        removes: list[str] = []
        adds: list[str] = []
        deletes: list[str] = []

        for operation in update.operations:
            field = self.convert_expr(
                Field(path=operation.field), attribute_names, attribute_values
            )
            if operation.op == UpdateOp.PUT:
                value = self.convert_expr(
                    operation.args[0], attribute_names, attribute_values
                )
                sets.append(f"{field} = {value}")
            elif operation.op == UpdateOp.INSERT:
                if field[-1] == "]":
                    value = self.convert_expr(
                        [operation.args[0]], attribute_names, attribute_values
                    )
                    if field.endswith("[-]"):
                        sets.append(
                            f"""{field[:-3]} = list_append({
                                field[:-3]}, {value})"""
                        )
                    elif field.endswith("[0]"):
                        sets.append(
                            f"""{field[:-3]} = list_append({
                                value}, {field[:-3]})"""
                        )
                    else:
                        raise BadRequestError(
                            "List update at intermediate index not supported"
                        )
                else:
                    value = self.convert_expr(
                        operation.args[0], attribute_names, attribute_values
                    )
                    sets.append(f"{field} = {value}")
            elif operation.op == UpdateOp.DELETE:
                removes.append(field)
            elif operation.op == UpdateOp.INCREMENT:
                val = operation.args[0]
                if operation.args[0] < 0:
                    val = -val
                    op = "-"
                else:
                    op = "+"
                value = self.convert_expr(
                    val, attribute_names, attribute_values
                )
                sets.append(f"{field} = {field} {op} {value}")
            elif operation.op == UpdateOp.MOVE:
                dest = self.convert_expr(
                    Field(path=operation.args[0].path),
                    attribute_names,
                    attribute_values,
                )
                sets.append(f"{dest} = {field}")
                removes.append(f"{field}")
            else:
                raise BadRequestError("Update operation not supported")

        expr = ""
        if len(sets) > 0:
            expr = f"{expr} SET {', '.join(e for e in sets)}"
        if len(removes) > 0:
            expr = f"{expr} REMOVE {', '.join(e for e in removes)}"
        if len(adds) > 0:
            expr = f"{expr} ADD {', '.join(e for e in adds)}"
        if len(deletes) > 0:
            expr = f"{expr} DELETE {', '.join(e for e in deletes)}"
        expr = expr.strip()
        return expr

    def convert_expr(
        self,
        expr: Expression | None,
        attribute_names: dict[str, str],
        attribute_values: dict[str, Any],
    ) -> str:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            var = f":v{len(attribute_values)}"
            if isinstance(expr, float):
                attribute_values[var] = Decimal(str(expr))
            else:
                attribute_values[var] = expr
            return var
        if isinstance(expr, Field):
            field = self.processor.resolve_field(expr.path)
            field = (
                field.replace("[", "/")
                .replace("]", "")
                .replace(".", "/")
                .rstrip("/")
            )
            splits = field.split("/")
            alias = ""
            for split in splits:
                if split == "-" or split.isdigit():
                    alias = f"{alias}[{split}]"
                else:
                    var = f"#f{len(attribute_names)}"
                    attribute_names[var] = split
                    alias = f"{alias}.{var}" if alias != "" else var
            return alias
        if isinstance(expr, Function):
            return self.convert_func(expr, attribute_names, attribute_values)
        if isinstance(expr, Comparison):
            lhs = f"""{self.convert_expr(
                    expr.lexpr, attribute_names, attribute_values)}"""
            op = expr.op.value
            rhs = None
            negate = False
            if expr.op == ComparisonOp.NEQ:
                op = "<>"
            if expr.op == ComparisonOp.BETWEEN and isinstance(
                expr.rexpr, list
            ):
                rhs = f"""{self.convert_expr(
                    expr.rexpr[0], attribute_names, attribute_values)} AND {
                        self.convert_expr(
                    expr.rexpr[1], attribute_names, attribute_values)}"""
            elif (
                expr.op == ComparisonOp.IN or expr.op == ComparisonOp.NIN
            ) and isinstance(expr.rexpr, list):
                lst = ", ".join(
                    self.convert_expr(i, attribute_names, attribute_values)
                    for i in expr.rexpr
                )
                rhs = f"({lst})"
                if expr.op == ComparisonOp.NIN:
                    op = "IN"
                    negate = True
            else:
                rhs = self.convert_expr(
                    expr.rexpr, attribute_names, attribute_values
                )
            if negate:
                return f"""NOT {lhs} {op} {rhs}"""
            return f"""{lhs} {op} {rhs}"""
        if isinstance(expr, And):
            return f"""({self.convert_expr(
                expr.lexpr, attribute_names, attribute_values)} AND {
                    self.convert_expr(
                expr.rexpr, attribute_names, attribute_values)})"""
        if isinstance(expr, Or):
            return f"""({self.convert_expr(
                expr.lexpr, attribute_names, attribute_values)} OR {
                    self.convert_expr(
                expr.rexpr, attribute_names, attribute_values)})"""
        if isinstance(expr, Not):
            return f"""NOT {self.convert_expr(
                expr.expr, attribute_names, attribute_values)}"""
        return str(expr)

    def convert_func(
        self,
        expr: Function,
        attribute_names: dict[str, str],
        attribute_values: dict[str, Any],
    ) -> str:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.IS_TYPE:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                type = args[1]
                if isinstance(type, str):
                    type = DocumentFieldType(type.lower())
                type_param = None
                if type == DocumentFieldType.STRING:
                    type_param = "S"
                elif type == DocumentFieldType.NUMBER:
                    type_param = "N"
                elif type == DocumentFieldType.BOOLEAN:
                    type_param = "BOOL"
                elif type == DocumentFieldType.OBJECT:
                    type_param = "M"
                elif type == DocumentFieldType.ARRAY:
                    type_param = "L"
                elif type == DocumentFieldType.NULL:
                    type_param = "NULL"
                value = self.convert_expr(
                    type_param, attribute_names, attribute_values
                )
                return f"attribute_type({field}, {value})"
            if name == StoreFunctionName.IS_DEFINED:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                return f"attribute_exists({field})"
            if name == StoreFunctionName.IS_NOT_DEFINED:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                return f"attribute_not_exists({field})"
            if name == StoreFunctionName.LENGTH:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                return f"size({field})"
            if name == StoreFunctionName.CONTAINS:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                value = self.convert_expr(
                    args[1], attribute_names, attribute_values
                )
                return f"contains({field}, {value})"
            if name == StoreFunctionName.STARTS_WITH:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                value = self.convert_expr(
                    args[1], attribute_names, attribute_values
                )
                return f"begins_with({field}, {value})"
            if name == StoreFunctionName.ARRAY_LENGTH:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                return f"size({field})"
            if name == StoreFunctionName.ARRAY_CONTAINS:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                value = self.convert_expr(
                    args[1], attribute_names, attribute_values
                )
                return f"contains({field}, {value})"
            if name == StoreFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_expr(
                    args[0], attribute_names, attribute_values
                )
                clauses = []
                for item in args[1]:
                    value = self.convert_expr(
                        item, attribute_names, attribute_values
                    )
                    clauses.append(f"contains({field}, {value})")
                return f"({str.join(' OR ', clauses)})"
        raise BadRequestError(f"Function {name} not recognized")

    def convert_select(
        self,
        select: Select | None,
        attribute_names: dict[str, str],
    ) -> tuple[str | None, str | None]:
        if select is None or len(select.terms) == 0:
            return "ALL_ATTRIBUTES", None
        return "SPECIFIC_ATTRIBUTES", ", ".join(
            self.convert_expr(Field(path=t.field), attribute_names, {})
            for t in select.terms
        )

    def convert_order_by(self, order_by: OrderBy | None) -> bool | None:
        if order_by is not None and len(order_by.terms) > 0:
            if order_by.terms[0].direction is not None:
                return order_by.terms[0].direction == OrderByDirection.ASC
            else:
                return True
        return None

    def convert_limit_offset(
        self, limit: int | None = None, offset: int | None = None
    ) -> tuple[int | None, int | None]:
        return limit, offset

    def _get_explicit_index(self, collection: str | None) -> str | None:
        if collection is not None and "." in collection:
            return collection.split(".")[-1]
        return None

    def _get_order_by_field(self, order_by: OrderBy | None) -> str | None:
        order_by_field = None
        if order_by is not None and len(order_by.terms) > 0:
            order_by_field = self.processor.resolve_field(
                order_by.terms[0].field
            )
        return order_by_field

    def _get_select_fields(self, select: Select | None) -> list[str]:
        select_fields = []
        if select is not None:
            for term in select.terms:
                field = self.processor.resolve_field(term.field)
                select_fields.append(field)
        return select_fields

    def _get_where_fields(self, expr: Expression | None) -> tuple:
        key_fields: dict = {}
        non_key_fields: list = []
        self._find_where_fields(expr, key_fields, non_key_fields, False)
        for f in non_key_fields:
            if f in key_fields:
                key_fields.pop(f)
        return key_fields, non_key_fields

    def _find_where_fields(
        self,
        expr: Expression | None,
        key_fields: dict[str, list],
        non_key_fields: list[str],
        non_key: bool,
    ):
        def try_add_key_field(field, index_types):
            field = self.processor.resolve_field(field)
            if field in key_fields:
                # If field already exists, add to non-key
                # DynamoDB supports only one comparison
                add_non_key_field(field)
            else:
                key_fields[field] = index_types

        def add_non_key_field(field):
            field = self.processor.resolve_field(field)
            if field not in non_key_fields:
                non_key_fields.append(field)

        if isinstance(expr, Comparison):
            field = None
            if isinstance(expr.lexpr, Field):
                field = expr.lexpr.path
            elif isinstance(expr.rexpr, Field):
                field = expr.rexpr.path
            if field is None:
                return
            if non_key:
                add_non_key_field(field)
                return
            if expr.op == ComparisonOp.EQ:
                try_add_key_field(
                    field,
                    [
                        IndexManager.INDEX_TYPE_HASH,
                        IndexManager.INDEX_TYPE_RANGE,
                    ],
                )
            elif expr.op in [
                ComparisonOp.LT,
                ComparisonOp.LTE,
                ComparisonOp.GT,
                ComparisonOp.GTE,
                ComparisonOp.BETWEEN,
            ]:
                try_add_key_field(field, [IndexManager.INDEX_TYPE_RANGE])
            else:
                if field is not None:
                    add_non_key_field(field)
        elif isinstance(expr, Function):
            if (
                expr.namespace == FunctionNamespace.BUILTIN
                and expr.name == StoreFunctionName.STARTS_WITH
                and isinstance(expr.args[0], Field)
                and not non_key
            ):
                try_add_key_field(
                    expr.args[0].path, [IndexManager.INDEX_TYPE_RANGE]
                )
            else:
                for arg in expr.args:
                    if isinstance(arg, Field):
                        add_non_key_field(arg.path)
        elif isinstance(expr, And):
            self._find_where_fields(
                expr.lexpr, key_fields, non_key_fields, non_key
            )
            self._find_where_fields(
                expr.rexpr, key_fields, non_key_fields, non_key
            )
        elif isinstance(expr, Or):
            self._find_where_fields(
                expr.lexpr, key_fields, non_key_fields, True
            )
            self._find_where_fields(
                expr.rexpr, key_fields, non_key_fields, True
            )
        elif isinstance(expr, Not):
            self._find_where_fields(
                expr.expr, key_fields, non_key_fields, True
            )

    def _split_key_filter_expression(
        self,
        expr: Expression,
        index: Index,
    ) -> tuple[Expression, Expression]:
        def is_key_field(field):
            field = self.processor.resolve_field(field)
            if field == index.hash_key or field == index.range_key:
                return True
            return False

        if isinstance(expr, Comparison):
            field = None
            if isinstance(expr.lexpr, Field):
                field = expr.lexpr.path
            elif isinstance(expr.rexpr, Field):
                field = expr.rexpr.path
            if is_key_field(field):
                return expr, Comparison(
                    lexpr=True, op=ComparisonOp.EQ, rexpr=True
                )
            return None, expr
        elif isinstance(expr, Function):
            for arg in expr.args:
                if isinstance(arg, Field):
                    if is_key_field(arg.path):
                        return expr, Comparison(
                            lexpr=True, op=ComparisonOp.EQ, rexpr=True
                        )
            return None, expr
        elif isinstance(expr, (And, Or)):
            (
                ke_lhs,
                fe_lhs,
            ) = self._split_key_filter_expression(expr.lexpr, index)

            (
                ke_rhs,
                fe_rhs,
            ) = self._split_key_filter_expression(expr.rexpr, index)

            ke: Expression = None
            if ke_lhs is not None and ke_rhs is not None:
                ke = And(lexpr=ke_lhs, rexpr=ke_rhs)
            elif ke_lhs is not None:
                ke = ke_lhs
            elif ke_rhs is not None:
                ke = ke_rhs

            fe: Expression = (
                And(lexpr=fe_lhs, rexpr=fe_rhs)
                if isinstance(expr, And)
                else Or(lexpr=fe_lhs, rexpr=fe_rhs)
            )
            return ke, fe
        elif isinstance(expr, Not):
            ke, fe = self._split_key_filter_expression(expr.expr, index)
            return ke, Not(expr=expr.expr)
        return None, expr


class IndexManager:
    INDEX_CATEGORY_MAIN = "main"
    INDEX_CATEGORY_LOCAL = "local"
    INDEX_CATEGORY_GLOBAL = "global"

    INDEX_TYPE_HASH = "hash"
    INDEX_TYPE_RANGE = "range"

    ACTION_QUERY = "query"
    ACTION_SCAN = "scan"

    table_name: str
    hash_key: str
    range_key: str | None
    indexes: list[Index]

    def __init__(self, table):
        self.indexes = []
        self.table_name = table.table_name
        self.hash_key, self.range_key = self._parse_key_schema(
            key_schema=table.key_schema
        )
        self._add_indexes(table=table)

    def __str__(self):
        return f"[{', '.join(str(i) for i in self.indexes)}]"

    def get_id_pk_field(self) -> tuple:
        if self.range_key is None:
            return self.hash_key, self.hash_key
        return self.range_key, self.hash_key

    def get_query_plan(
        self,
        order_by_field: str | None,
        select_fields: list[str],
        where_key_fields: dict[str, list[str]],
        where_non_key_fields: list[str],
        explicit_index: str | None,
    ) -> QueryPlan:
        candidate_indexes = self._get_order_by_candidate_indexes(
            order_by_field, self.indexes
        )
        if order_by_field is not None and len(candidate_indexes) == 0:
            raise BadRequestError(
                f"No index found to order by {order_by_field}"
            )

        candidate_indexes = self._get_select_candidate_indexes(
            select_fields, candidate_indexes
        )
        if order_by_field is not None and len(candidate_indexes) == 0:
            raise BadRequestError(
                f"""No index found to order by {
                order_by_field} and get the select fields"""
            )

        candidate_indexes = self._get_where_candidate_indexes(
            where_key_fields, where_non_key_fields, candidate_indexes
        )
        if order_by_field is not None and len(candidate_indexes) == 0:
            raise BadRequestError(
                f"""No index found to order by {
                order_by_field} and filter by where condition"""
            )

        action = IndexManager.ACTION_SCAN
        index = self.indexes[0]
        if explicit_index is not None:
            match = False
            for candidate_index in candidate_indexes:
                if candidate_index.name == explicit_index:
                    action = IndexManager.ACTION_QUERY
                    index = candidate_index
                    match = True
                    break
            if order_by_field is not None and not match:
                raise BadRequestError(
                    f"""Provided index does not support the order by {
                    order_by_field} query"""
                )
            if not match:
                for db_index in self.indexes:
                    if db_index.name == explicit_index:
                        action = IndexManager.ACTION_SCAN
                        index = db_index
        else:
            if len(candidate_indexes) > 0:
                action = IndexManager.ACTION_QUERY
                index = candidate_indexes[0]

        query_plan = QueryPlan()
        query_plan.action = action
        query_plan.index = index
        return query_plan

    def _get_order_by_candidate_indexes(
        self,
        order_by_field: str | None,
        current_candidate_indexes: list[Index],
    ) -> list[Index]:
        candidate_indices = []
        for index in current_candidate_indexes:
            if order_by_field is None or index.range_key == order_by_field:
                candidate_indices.append(index)
        return candidate_indices

    def _get_select_candidate_indexes(
        self, select_fields: list[str], current_candidate_indexes: list[Index]
    ) -> list[Index]:
        candidate_indices: list[Index] = []
        for index in current_candidate_indexes:
            if index.projection_type == "ALL":
                # rank higher
                candidate_indices.insert(0, index)
            else:
                if len(select_fields) > 0:
                    match = True
                    for select_field in select_fields:
                        if select_field not in index.projection_fields:
                            match = False
                    if match:
                        # rank higher
                        candidate_indices.insert(0, index)
                    elif index.category == IndexManager.INDEX_CATEGORY_LOCAL:
                        # rank lower since will require extra reads
                        candidate_indices.append(index)

        return candidate_indices

    def _get_where_candidate_indexes(
        self,
        where_key_fields: dict[str, list[str]],
        where_non_key_fields: list[str],
        current_candidate_indexes: list[Index],
    ) -> list[Index]:
        candidate_indexes: list[Index] = []
        for index in reversed(current_candidate_indexes):
            if (
                index.hash_key in where_key_fields
                and IndexManager.INDEX_TYPE_HASH
                in where_key_fields[index.hash_key]
            ):
                if index.range_key is not None:
                    if index.range_key in where_key_fields:
                        # rank higher
                        candidate_indexes.insert(0, index)
                    if index.range_key not in where_non_key_fields:
                        # rank lower
                        candidate_indexes.append(index)
                else:
                    # rank lower
                    candidate_indexes.append(index)

        return candidate_indexes

    def _add_keys_to_projection(
        self, keys: list[str | None], projection_fields: list[str]
    ):
        for key in keys:
            if key is not None and key not in projection_fields:
                projection_fields.append(key)

    def _add_indexes(self, table):
        self.indexes.append(self._get_main_index())

        for index_category in [
            IndexManager.INDEX_CATEGORY_LOCAL,
            IndexManager.INDEX_CATEGORY_GLOBAL,
        ]:
            db_indexes = None
            if index_category == IndexManager.INDEX_CATEGORY_GLOBAL:
                db_indexes = table.local_secondary_indexes
            elif index_category == IndexManager.INDEX_CATEGORY_LOCAL:
                db_indexes = table.global_secondary_indexes
            if db_indexes is not None:
                for db_index in db_indexes:
                    index = self._parse_index(db_index)
                    index.category = index_category
                    if index.projection_type != "ALL":
                        self._add_keys_to_projection(
                            [
                                self.hash_key,
                                self.range_key,
                                index.hash_key,
                                index.range_key,
                            ],
                            index.projection_fields,
                        )
                    self.indexes.append(index)

    def _get_main_index(self) -> Index:
        index = Index()
        index.name = "$main"
        index.category = IndexManager.INDEX_CATEGORY_MAIN
        index.hash_key = self.hash_key
        index.range_key = self.range_key
        index.projection_type = "ALL"
        index.projection_fields = []
        return index

    def _parse_index(self, db_index: dict) -> Index:
        index = Index()
        index.name = f"{self.table_name}.{db_index['IndexName']}"
        index.hash_key, index.range_key = self._parse_key_schema(
            db_index["KeySchema"]
        )
        (
            index.projection_type,
            index.projection_fields,
        ) = self._parse_projection(db_index["Projection"])
        return index

    def _parse_key_schema(self, key_schema: list[dict]) -> tuple:
        hash_key = None
        range_key = None
        if len(key_schema) == 2:
            if key_schema[0]["KeyType"] == "HASH":
                hash_key = key_schema[0]["AttributeName"]
                range_key = key_schema[1]["AttributeName"]
            elif key_schema[0]["KeyType"] == "RANGE":
                hash_key = key_schema[1]["AttributeName"]
                range_key = key_schema[0]["AttributeName"]
        else:
            hash_key = key_schema[0]["AttributeName"]
        return hash_key, range_key

    def _parse_projection(self, projection: dict) -> tuple:
        projection_type = projection["ProjectionType"]
        projection_fields = []
        if projection_type == "INCLUDE":
            projection_fields = projection["NonKeyAttributes"]
        return projection_type, projection_fields


class Index:
    name: str | None
    hash_key: str
    range_key: str | None
    projection_type: str
    projection_fields: list[str]
    category: str

    def __repr__(self):
        return str(self.__dict__)


class QueryPlan:
    action: str
    index: Index


class Query:
    action: str = IndexManager.ACTION_SCAN
    index_name: str | None = None
    select: str | None = None
    projection_expression: str | None = None
    key_condition_expression: str | None = None
    filter_expression: str | None = None
    scan_index_forward: bool | None = None
    limit: int | None = None
    offset: int | None = None
    expression_attribute_values: dict[str, Any] = {}
    expression_attribute_names: dict[str, str] = {}

    def __repr__(self) -> str:
        return str(self.__dict__)


class DynamoDBCollection:
    client: Any
    converter: OperationConverter
    processor: ItemProcessor
    helper: Any

    def __init__(
        self,
        client: Any,
        helper_type: Any,
        id_map_field: str | None,
        pk_map_field: str | None,
        etag_embed_field: str | None,
        suppress_fields: list[str] | None,
    ):
        self.client = client
        index_manager = IndexManager(table=client)
        id_embed_field, pk_embed_field = index_manager.get_id_pk_field()
        self.processor = ItemProcessor(
            id_embed_field=id_embed_field,
            pk_embed_field=pk_embed_field,
            etag_embed_field=etag_embed_field,
            id_map_field=id_map_field,
            pk_map_field=pk_map_field,
            local_etag=True,
            suppress_fields=suppress_fields,
        )
        self.converter = OperationConverter(self.processor, index_manager)
        self.helper = helper_type(client)



from datetime import datetime

from verse.core import DataModel

from ._constants import DEFAULT_LABEL


class ConfigKey(DataModel):
    """Config key."""

    id: str
    """Config id.
    """

    label: str | None = DEFAULT_LABEL
    """Config label.
    """


class ConfigProperties(DataModel):
    """Config properties."""

    updated_time: datetime | None = None
    """Config updated time.
    """


class ConfigItem(DataModel):
    """Config item."""

    key: ConfigKey
    """Config key.
    """

    value: str | None = None
    """Config value.
    """

    metadata: dict | None = None
    """Config metadata.
    """

    properties: ConfigProperties | None = None
    """Config properties.
    """


class ConfigList(DataModel):
    """Config list."""

    items: list[ConfigItem]
    """List of config items.
    """



import json


def get_credentials(
    service_account_info,
    service_account_file,
    access_token,
):
    from google.oauth2 import service_account

    if service_account_info is not None:
        service_account_info = (
            json.loads(service_account_info)
            if isinstance(service_account_info, str)
            else service_account_info
        )
        return service_account.Credentials.from_service_account_info(
            service_account_info
        )
    elif service_account_file is not None:
        return service_account.Credentials.from_service_account_file(
            service_account_file
        )
    elif access_token is not None:
        from google.auth.credentials import Credentials

        return Credentials(access_token)
    return None



import json
import uuid
from typing import Any

from verse.ql import Update, Value

from ._attributes import Attribute, KeyAttribute, SpecialAttribute


def root(attribute: str) -> str:
    return f"${attribute}"


def unroot(attribute: str) -> str:
    if attribute[0] == "$":
        return attribute[1:]
    return attribute


class ItemProcessor:
    id_embed_field: str | None
    pk_embed_field: str | None
    etag_embed_field: str | None
    id_map_field: str | None
    pk_map_field: str | None
    local_etag: bool | None
    suppress_fields: list[str] | None
    field_types: dict | None

    def __init__(
        self,
        id_embed_field: str | None = None,
        pk_embed_field: str | None = None,
        etag_embed_field: str | None = None,
        id_map_field: str | None = None,
        pk_map_field: str | None = None,
        local_etag: bool | None = False,
        suppress_fields: list[str] | None = None,
        field_types: dict | None = None,
        **kwargs,
    ):
        self.id_embed_field = id_embed_field
        self.pk_embed_field = pk_embed_field
        self.etag_embed_field = etag_embed_field
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field
        self.local_etag = local_etag
        self.suppress_fields = suppress_fields
        self.field_types = field_types

    def suppress_fields_if_needed(self, value: dict) -> dict:
        if self.suppress_fields is not None:
            return {
                k: v for k, v in value.items() if k not in self.suppress_fields
            }
        return value

    def add_embed_fields(
        self,
        value: dict,
        key: Value | None = None,
    ) -> dict:
        value_copy = None
        if key is not None:
            value_copy = value.copy()
            if self.id_embed_field is not None:
                value_copy[self.id_embed_field] = self.get_id_from_key(key=key)
            if (
                self.pk_embed_field is not None
                and self.id_embed_field != self.pk_embed_field
            ):
                value_copy[self.pk_embed_field] = self.get_pk_from_key(key=key)
            if (
                self.id_map_field is not None
                and self.id_map_field not in value
            ):
                value_copy[self.id_map_field] = self.get_id_from_key(key=key)
            if (
                self.pk_map_field is not None
                and self.pk_map_field not in value
            ):
                value_copy[self.pk_map_field] = self.get_pk_from_key(key=key)
        elif self.id_map_field is not None or self.pk_map_field is not None:
            value_copy = value.copy()
            if (
                self.id_embed_field is not None
                and self.id_map_field is not None
            ):
                value_copy[self.id_embed_field] = value[self.id_map_field]
            if (
                self.pk_embed_field is not None
                and self.pk_map_field is not None
                and self.id_embed_field != self.pk_embed_field
            ):
                value_copy[self.pk_embed_field] = value[self.pk_map_field]
        if self.local_etag and self.etag_embed_field is not None:
            if value_copy is None:
                value_copy = value.copy()
            value_copy[self.etag_embed_field] = self.generate_etag()
        if value_copy is not None:
            return value_copy
        return value

    def get_id_from_key(self, key: Value) -> Any:
        if isinstance(key, dict):
            if self.id_embed_field is not None and self.id_embed_field in key:
                return key[self.id_embed_field]
            if self.id_map_field is not None and self.id_map_field in key:
                return key[self.id_map_field]
            if SpecialAttribute.ID in key:
                return key[SpecialAttribute.ID]
            if KeyAttribute.ID in key:
                return key[KeyAttribute.ID]
        elif isinstance(key, (str | int | float | bool)):
            return key
        return None

    def get_pk_from_key(self, key: Value) -> Any:
        if isinstance(key, dict):
            if self.pk_embed_field is not None and self.pk_embed_field in key:
                return key[self.pk_embed_field]
            if self.pk_map_field is not None and self.pk_map_field in key:
                return key[self.pk_map_field]
            if SpecialAttribute.PK in key:
                return key[SpecialAttribute.PK]
            if KeyAttribute.PK in key:
                return key[KeyAttribute.PK]
        elif isinstance(key, (str | int | float | bool)):
            return key
        return None

    def get_id_pk_from_key(self, key: Value) -> tuple:
        return self.get_id_from_key(key=key), self.get_pk_from_key(key=key)

    def get_id_from_value(self, value: dict) -> Any:
        if self.id_embed_field is not None and self.id_embed_field in value:
            return value[self.id_embed_field]
        if self.id_map_field is not None and self.id_map_field in value:
            return value[self.id_map_field]
        return None

    def get_pk_from_value(self, value: dict) -> Any:
        if self.pk_embed_field is not None and self.pk_embed_field in value:
            return value[self.pk_embed_field]
        if self.pk_map_field is not None and self.pk_map_field in value:
            return value[self.pk_map_field]
        return None

    def get_id_pk_from_value(self, value: dict) -> tuple:
        return self.get_id_from_value(value=value), self.get_pk_from_value(
            value=value
        )

    def get_key_from_key(self, key: Value) -> dict:
        rkey = {}
        id = self.get_id_from_key(key=key)
        if id is not None and self.id_embed_field is not None:
            rkey[self.id_embed_field] = id
        elif id is not None and self.id_map_field is not None:
            rkey[self.id_map_field] = id
        pk = self.get_pk_from_key(key=key)
        if (
            pk is not None
            and self.pk_embed_field is not None
            and self.id_embed_field != self.pk_embed_field
        ):
            rkey[self.pk_embed_field] = pk
        elif (
            pk is not None
            and self.pk_map_field is not None
            and self.id_map_field != self.pk_map_field
        ):
            rkey[self.pk_map_field] = pk
        return rkey

    def get_key_from_value(self, value: dict) -> dict:
        key = {}
        id = self.get_id_from_value(value=value)
        if id is not None and self.id_embed_field is not None:
            key[self.id_embed_field] = id
        elif id is not None and self.id_map_field is not None:
            key[self.id_map_field] = id
        pk = self.get_pk_from_value(value=value)
        if (
            pk is not None
            and self.pk_embed_field is not None
            and self.id_embed_field != self.pk_embed_field
        ):
            key[self.pk_embed_field] = pk
        elif (
            pk is not None
            and self.pk_map_field is not None
            and self.id_map_field != self.pk_map_field
        ):
            key[self.pk_map_field] = pk
        return key

    def get_normalized_key_from_key(self, key: Value) -> dict:
        rkey = {}
        id = self.get_id_from_key(key=key)
        if id is not None:
            rkey[KeyAttribute.ID] = id
        pk = self.get_pk_from_key(key=key)
        if pk is not None:
            rkey[KeyAttribute.PK] = pk
        return rkey

    def get_normalized_key_from_value(self, value: dict) -> dict:
        key = {}
        id = self.get_id_from_value(value=value)
        if id is not None:
            key[KeyAttribute.ID] = id
        pk = self.get_pk_from_value(value=value)
        if pk is not None:
            key[KeyAttribute.PK] = pk
        return key

    def get_etag_from_value(self, value: dict) -> str | None:
        if (
            self.etag_embed_field is not None
            and self.etag_embed_field in value
        ):
            return value[self.etag_embed_field]
        return None

    def needs_local_etag(self) -> bool:
        if self.local_etag is None:
            return False
        return self.local_etag and self.etag_embed_field is not None

    def add_etag_update(self, update: Update, etag: str) -> Update:
        if self.local_etag and self.etag_embed_field is not None:
            update.put(self.etag_embed_field, etag)
        return update

    def generate_etag(self) -> str:
        return str(uuid.uuid4())

    def convert_type(self, field: str, value: Any) -> Any:
        if self.field_types is None:
            return value
        if field not in self.field_types:
            return value
        type = self.field_types[field]
        if type == "object":
            return json.loads(value)
        if type == "array":
            return json.loads(value)
        if type == "integer":
            return int(value)
        if type == "float":
            return float(value)
        if type == "number":
            try:
                return int(value)
            except ValueError:
                return float(value)
        if type == "string":
            return str(value)
        if type == "boolean":
            if isinstance(value, str):
                value_lower = value.lower()
                if value_lower in ["0", "false"]:
                    return False
                elif value_lower in ["1", "true"]:
                    return True
            return bool(value)
        if type == "string|null":
            if value == "null":
                return None
            return str(value)
        return value

    def resolve_field(self, field: str) -> str:
        if field == SpecialAttribute.ID and self.id_map_field is not None:
            return self.id_map_field
        if field == SpecialAttribute.PK and self.pk_map_field is not None:
            return self.pk_map_field
        if field == SpecialAttribute.ID and self.id_embed_field is not None:
            return self.id_embed_field
        if field == SpecialAttribute.PK and self.pk_embed_field is not None:
            return self.pk_embed_field
        if (
            field == SpecialAttribute.ETAG
            and self.etag_embed_field is not None
        ):
            return self.etag_embed_field
        return field

    def resolve_attribute_field(self, field: str, attribute: str) -> str:
        if field.startswith(f"${attribute}."):
            return field.replace(f"${attribute}.", "")
        return field

    def resolve_metadata_field(self, field: str) -> str:
        return self.resolve_attribute_field(field, Attribute.METADATA)

    def resolve_root_field(self, field: str) -> str:
        if field == SpecialAttribute.ID:
            return "key.id"
        if field == SpecialAttribute.LABEL:
            return "key.label"
        if field == SpecialAttribute.SCORE:
            return "properties.score"
        if field[0] == "$":
            return field[1:]
        return f"value.{field}"



from verse.core import Operation, Provider

from ._operation_parser import StoreOperationParser


class StoreProvider(Provider):
    def get_op_parser(
        self, operation: Operation | None
    ) -> StoreOperationParser:
        return StoreOperationParser(operation)



"""
Secret Store on top of SQLite.
"""

__all__ = ["SQLite"]

import json
import sqlite3
import uuid
from datetime import datetime
from typing import Any

from verse.core import Context, Operation, Response
from verse.core.exceptions import NotFoundError, PreconditionFailedError
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreOperation,
    StoreProvider,
)

from .._models import (
    SecretItem,
    SecretKey,
    SecretList,
    SecretProperties,
    SecretVersion,
)


class SQLite(StoreProvider):
    database: str
    nparams: dict[str, Any]

    _client: Any
    _processor: ItemProcessor

    def __init__(
        self,
        database: str = ":memory:",
        nparams: dict[str, Any] = dict(),
        **kwargs
    ):
        """Initialize.

        Args:
            database:
                SQLite database. Defaults to ":memory:".
            nparams:
                Native parameters to SQLite client.
        """
        self.database = database
        self.nparams = nparams

        self._client = None
        self._processor = ItemProcessor()

    def init(
        self,
        context: Context | None = None,
    ) -> None:
        if self._client is not None:
            return

        self._client = sqlite3.connect(
            database=self.database,
            check_same_thread=False,
            **self.nparams,
        )
        cursor = self._client.cursor()
        rows = cursor.execute(
            """SELECT name FROM sqlite_master WHERE type='table'
                AND (name=? or name=?)""",
            ("items", "versions"),
        ).fetchall()
        if len(rows) == 0:
            self._create_tables(cursor)
        cursor.close()

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        client = self._client
        op_parser = self.get_op_parser(operation)
        processor = self._processor
        nresult: Any = None
        result: Any = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            query_version = version
            cursor = client.cursor()
            try:
                if query_version is None:
                    nresult = cursor.execute(
                        """SELECT version FROM items
                            WHERE id = ? AND deleted = 0""",
                        (id,),
                    ).fetchone()
                    if nresult is None:
                        raise NotFoundError
                    (query_version,) = nresult
                nresult = cursor.execute(
                    """SELECT value FROM versions WHERE id = ?
                        AND version = ?""",
                    (id, query_version),
                ).fetchone()
                (value,) = nresult
                result = SecretItem(
                    key=SecretKey(id=id, version=query_version),
                    value=value,
                )
            finally:
                cursor.close()
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            cursor = client.cursor()
            try:
                nresult = cursor.execute(
                    """SELECT metadata, created_time FROM items
                        WHERE id = ? AND deleted = 0""",
                    (id,),
                ).fetchone()
                if nresult is None:
                    raise NotFoundError
                (metadata, created_time) = nresult
                result = SecretItem(
                    key=SecretKey(id=id),
                    metadata=json.loads(metadata),
                    properties=SecretProperties(
                        created_time=Helper.convert_to_datetime(created_time)
                    ),
                )
            finally:
                cursor.close()
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            cursor = client.cursor()
            try:
                versions: list = []
                nresult = cursor.execute(
                    """SELECT id FROM items
                        WHERE id = ? AND deleted = 0""",
                    (id,),
                ).fetchone()
                if nresult is None:
                    raise NotFoundError
                nresult = cursor.execute(
                    """SELECT version, created_time FROM versions
                        WHERE id = ? ORDER BY created_time DESC""",
                    (id,),
                ).fetchall()
                if len(nresult) == 0:
                    raise NotFoundError
                for item in nresult:
                    (version, created_time) = item
                    versions.append(
                        SecretVersion(
                            version=str(version),
                            created_time=Helper.convert_to_datetime(
                                created_time
                            ),
                        )
                    )
                result = SecretItem(key=SecretKey(id=id), versions=versions)
            finally:
                cursor.close()
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            metadata = op_parser.get_metadata()
            exists = op_parser.get_where_exists()
            version = str(uuid.uuid4())
            created_time = datetime.now()
            cursor = client.cursor()
            insert = True
            if exists is None or exists is False:
                try:
                    nresult = cursor.execute(
                        """INSERT INTO items (id, version, metadata,
                            created_time, deleted) VALUES (?, ?, ?, ?, ?)""",
                        (
                            id,
                            version,
                            json.dumps(
                                metadata if metadata is not None else {}
                            ),
                            created_time,
                            0,
                        ),
                    )
                except sqlite3.IntegrityError:
                    insert = False
                    if exists is False:
                        raise PreconditionFailedError
            if exists is False or exists is True or insert is False:
                nresult = cursor.execute(
                    """UPDATE items SET version = ?,
                        metadata = ? WHERE id = ? AND deleted = 0
                        RETURNING id""",
                    (
                        version,
                        json.dumps(metadata if metadata is not None else {}),
                        id,
                    ),
                ).fetchone()
                if nresult is None and exists is True:
                    raise PreconditionFailedError
            nresult = cursor.execute(
                """INSERT INTO versions (id, version, value,
                    created_time) VALUES (?, ?, ?, ?)""",
                (id, version, value, created_time),
            )
            cursor.close()
            result = SecretItem(key=SecretKey(id=id, version=version))
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            value = op_parser.get_updated_value()
            version = str(uuid.uuid4())
            created_time = datetime.now()
            cursor = client.cursor()
            try:
                nresult = cursor.execute(
                    """UPDATE items SET version = ?
                        WHERE id = ? AND deleted = 0 RETURNING *""",
                    (version, id),
                ).fetchone()
                if nresult is None:
                    raise NotFoundError
                nresult = cursor.execute(
                    """INSERT INTO versions (id, version, value,
                        created_time) VALUES (?, ?, ?, ?)""",
                    (id, version, value, created_time),
                )
                result = SecretItem(key=SecretKey(id=id, version=version))
            finally:
                cursor.close()
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            cursor = client.cursor()
            try:
                nresult = cursor.execute(
                    """UPDATE items SET metadata = ?
                        WHERE id = ? AND deleted = 0 RETURNING id""",
                    (json.dumps(metadata if metadata is not None else {}), id),
                ).fetchone()
                if nresult is None:
                    raise NotFoundError
                result = SecretItem(
                    key=SecretKey(id=id),
                    metadata=metadata if metadata is not None else {},
                )
            finally:
                cursor.close()
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            cursor = client.cursor()
            try:
                nresult = cursor.execute(
                    "DELETE FROM items WHERE id = ? RETURNING id",
                    (id,),
                ).fetchone()
                if nresult is None:
                    raise NotFoundError
                nresult = cursor.execute(
                    "DELETE FROM versions WHERE id = ?", (id,)
                )
                result = None
            finally:
                cursor.close()
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            cursor = client.cursor()
            items = []
            try:
                nresult = cursor.execute(
                    """SELECT id, metadata, created_time FROM items
                        WHERE deleted=0 ORDER BY id"""
                )
                for item in nresult:
                    (id, metadata, created_time) = item
                    items.append(
                        SecretItem(
                            key=SecretKey(id=id),
                            metadata=json.loads(metadata),
                            properties=SecretProperties(
                                created_time=Helper.convert_to_datetime(
                                    created_time
                                )
                            ),
                        )
                    )
                items = QueryProcessor.query_items(
                    items=items,
                    select=op_parser.get_select(),
                    where=op_parser.get_where(),
                    order_by=op_parser.get_order_by(),
                    limit=op_parser.get_limit(),
                    offset=op_parser.get_offset(),
                    field_resolver=processor.resolve_root_field,
                )
                result = SecretList(items=items)
            finally:
                cursor.close()
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            cursor = client.cursor()
            items = []
            try:
                nresult = cursor.execute(
                    """SELECT id, metadata, created_time FROM items
                        WHERE deleted=0 ORDER BY id"""
                )
                for item in nresult:
                    (id, metadata, created_time) = item
                    items.append(
                        SecretItem(
                            key=SecretKey(id=id),
                            metadata=json.loads(metadata),
                            properties=SecretProperties(
                                created_time=Helper.convert_to_datetime(
                                    created_time
                                )
                            ),
                        )
                    )
                result = QueryProcessor.count_items(
                    items=items,
                    where=op_parser.get_where(),
                    field_resolver=processor.resolve_root_field,
                )
            finally:
                cursor.close()
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            pass
        else:
            return super().run(operation)
        return Response(result=result, native=dict(result=nresult))

    def _create_tables(self, cursor: Any, **kwargs: Any) -> None:
        cursor.execute(
            """CREATE TABLE items (id TEXT PRIMARY KEY,
                version TEXT, metadata JSON, created_time TIMESTAMP,
                deleted BOOLEAN, deleted_time TIMESTAMP, UNIQUE(id))"""
        )
        cursor.execute(
            """CREATE TABLE versions (id TEXT, version TEXT,
                value TEXT, created_time TIMESTAMP,
                PRIMARY KEY (id, version),
                UNIQUE(id, version))"""
        )


class Helper:
    @staticmethod
    def convert_to_datetime(timestamp: str):
        return datetime.fromisoformat(timestamp).astimezone()



from verse.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from verse.internal.storage_core import StoreFunction

from ._models import (
    SparseVectorIndex,
    VectorCollectionConfig,
    VectorIndex,
    VectorIndexMetric,
    VectorIndexStructure,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorValue,
)
from .component import VectorOperation, VectorStore

__all__ = [
    "SparseVectorIndex",
    "VectorCollectionConfig",
    "VectorIndex",
    "VectorIndexMetric",
    "VectorIndexStructure",
    "VectorItem",
    "VectorKey",
    "VectorList",
    "VectorOperation",
    "VectorProperties",
    "VectorStore",
    "VectorValue",
    "StoreFunction",
    "BadRequestError",
    "ConflictError",
    "NotFoundError",
    "PreconditionFailedError",
]



"""
Vector Store on Chroma.
"""

from __future__ import annotations

__all__ = ["Chroma"]

from typing import Any

import chromadb
import chromadb.config

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import (
    ItemProcessor,
    ParameterParser,
    QueryProcessor,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    Select,
)

from .._helper import get_collection_config, is_metadata_field, is_value_field
from .._models import (
    VectorCollectionConfig,
    VectorIndexMetric,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorSearchArgs,
    VectorValue,
)

DEFAULT_LIMIT = 100


class Chroma(StoreProvider):
    type: str | None
    collection: str | None
    dimension: int | dict[str, int]
    tenant: str
    database: str
    path: str
    host: str
    port: int
    ssl: bool
    auth_provider: str | None
    auth_credentials: str | None
    nparams: dict[str, Any]

    _init: bool = False
    _database_client: chromadb.ClientAPI
    _processor: ItemProcessor
    _collection_cache: dict[str, ChromaCollection]

    def __init__(
        self,
        type: str | None = "memory",
        collection: str | None = None,
        dimension: int | dict[str, int] = 4,
        tenant: str = "default_tenant",
        database: str = "default_database",
        path: str = "./chroma",
        host: str = "localhost",
        port: int = 8000,
        ssl: bool = False,
        auth_provider: str | None = None,
        auth_credentials: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            type:
                Chroma client type. "memory", "persistant", or "http".
                Defaults to "memory".
            collection:
                Chroma collection name mapped to vector store collection.
            dimension:
                Vector dimension.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the dimension.
            tenant:
                Chroma tenant name. Defaults to "default_tenant".
            database:
                Chroma database name. Defaults to "default_database".
            path:
                Path to persistent database. Defaults to "./chroma".
            host:
                Host for http client. Defaults to "localhost".
            port:
                Port for http client. Defaults to 8000.
            ssl:
                SSL option for http client. Defaults to False.
            auth_provider:
                Auth provider for http client. Defaults to None.
            auth_credentials:
                Auth credentials for http client. Defaults to None.
            nparams:
                Native parameters to chroma client.
        """
        self.type = type
        self.collection = collection
        self.dimension = dimension
        self.tenant = tenant
        self.database = database
        self.path = path
        self.host = host
        self.port = port
        self.ssl = ssl
        self.auth_provider = auth_provider
        self.auth_credentials = auth_credentials
        self.nparams = nparams

        self._init = False
        self._processor = ItemProcessor()
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._init:
            return

        if self.type == "memory":
            self._database_client = chromadb.Client(
                tenant=self.tenant,
                database=self.database,
                **self.nparams,
            )
        elif self.type == "persistent":
            self._database_client = chromadb.PersistentClient(
                path=self.path,
                tenant=self.tenant,
                database=self.database,
                **self.nparams,
            )
        elif self.type == "http":
            settings = None
            if self.auth_provider is not None:
                cred = self.auth_credentials
                settings = chromadb.config.Settings(
                    chroma_client_auth_provider=self.auth_provider,
                    chroma_client_auth_credentials=cred,
                )
            if settings is None:
                self._database_client = chromadb.HttpClient(
                    host=self.host,
                    port=self.port,
                    ssl=self.ssl,
                    **self.nparams,
                )
            else:
                self._database_client = chromadb.HttpClient(
                    host=self.host,
                    port=self.port,
                    ssl=self.ssl,
                    settings=settings,
                    **self.nparams,
                )
        else:
            raise BadRequestError(f"Type {self.type} is not supported")
        self._init = True

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            or op_parser.op_equals(StoreOperation.TRANSACT)
            else op_parser.get_collection_name()
        )
        collection = (
            collection_name if collection_name is not None else self.collection
        )
        return collection

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> ChromaCollection | None:
        if op_parser.is_resource_op():
            return None
        collection_name = self._get_collection_name(op_parser)
        if collection_name is None:
            raise BadRequestError("Collection name must be specified")
        if collection_name in self._collection_cache:
            return self._collection_cache[collection_name]
        client = self._database_client.get_collection(collection_name)
        dimension = ParameterParser.get_collection_parameter(
            self.dimension, collection_name
        )
        col = ChromaCollection(
            client,
            dimension,
        )
        self._collection_cache[collection_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            Validator.validate_transact(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._database_client),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: ChromaCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if collection is not None:
            converter = collection.converter
            client = collection.client
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args: dict[Any, Any] | None = {
                "collection": self._get_collection_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_id_as_str())
            call = NCall(client.get, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args = converter.convert_put(
                op_parser.get_id_as_str(),
                op_parser.get_vector_value(),
                op_parser.get_metadata(),
            )
            call = NCall(client.upsert, args, nargs)
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            args = converter.convert_update_value(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_updated_value(),
            )
            call = NCall(client.update, args, nargs)
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            args = converter.convert_update_metadata(
                id=op_parser.get_id_as_str(),
                metadata=op_parser.get_updated_metadata(),
            )
            call = NCall(client.update, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = converter.convert_delete(
                op_parser.get_id_as_str_or_none(), op_parser.get_where()
            )
            call = NCall(
                client.delete,
                args,
                nargs,
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                search=op_parser.get_search_as_function(),
                where=op_parser.get_where(),
                select=op_parser.get_select(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            call = NCall(client.query, args, nargs)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(client.count, args, nargs)
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            args, func = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "get":
                call = NCall(client.get, args, nargs)
            elif func == "put":
                call = NCall(client.upsert, args, nargs)
            elif func == "delete":
                call = NCall(client.delete, args, nargs)
            state = {"func": func}
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collection: ChromaCollection | None,
    ) -> Any:
        if collection is not None:
            processor = collection.processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = [c.name for c in nresult]
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            if len(nresult["ids"]) == 0:
                raise NotFoundError
            result = self._convert_to_items(nresult, processor)[0]
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            result = VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
                metadata=op_parser.get_updated_metadata(),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items: list[VectorItem] = self._convert_to_items(
                nresult, processor, is_query=True
            )
            items = QueryProcessor.order_items(
                items, op_parser.get_order_by(), processor.resolve_root_field
            )
            limit = op_parser.get_limit()
            items = QueryProcessor.limit_items(
                items,
                limit if limit is not None else DEFAULT_LIMIT,
                op_parser.get_offset(),
            )
            result = VectorList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            result = []
            func = state["func"] if state is not None else None
            if func == "get":
                result = self._convert_to_items(nresult, processor)
            elif func == "put":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(
                        VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
                    )
            elif func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result

    def _convert_to_items(
        self, nitems: Any, processor: ItemProcessor, is_query: bool = False
    ) -> list[VectorItem]:
        items = []
        count = len(nitems["ids"][0]) if is_query else len(nitems["ids"])
        for i in range(0, count):
            vector = None
            content = None
            metadata = None
            if is_query:
                key = VectorKey(id=nitems["ids"][0][i])
            else:
                key = VectorKey(id=nitems["ids"][i])
            if "metadatas" in nitems and nitems["metadatas"]:
                if is_query:
                    metadata = nitems["metadatas"][0][i]
                else:
                    metadata = nitems["metadatas"][i]
            if "embeddings" in nitems and nitems["embeddings"] is not None:
                if is_query:
                    vector = nitems["embeddings"][0][i]
                else:
                    vector = nitems["embeddings"][i]
            if "documents" in nitems and nitems["documents"]:
                if is_query:
                    content = nitems["documents"][0][i]
                else:
                    content = nitems["documents"][i]
            value = VectorValue(
                vector=vector,
                content=content,
            )
            properties = None
            if "distances" in nitems and nitems["distances"]:
                if is_query:
                    score = nitems["distances"][0][i]
                else:
                    score = nitems["distances"][i]
                properties = VectorProperties(score=1 - score)
            items.append(
                VectorItem(
                    key=key,
                    value=value,
                    metadata=metadata,
                    properties=properties,
                )
            )
        return items


class ResourceHelper:
    database_client: chromadb.ClientAPI

    def __init__(self, database_client: Any):
        self.database_client = database_client

    def create_collection(
        self,
        collection: str,
        config: VectorCollectionConfig | None,
        nargs: Any,
    ):
        metric_mapping = {
            VectorIndexMetric.DOTPRODUCT: "ip",
            VectorIndexMetric.COSINE: "cosine",
            VectorIndexMetric.EUCLIDEAN: "l2",
        }
        vector_index = config.vector_index if config is not None else None
        if vector_index is not None:
            metric = metric_mapping[vector_index.metric]
        else:
            metric = "ip"

        args = {
            "name": collection,
            "metadata": {"hnsw:space": metric},
            "get_or_create": True,
        }
        NCall(self.database_client.create_collection, args, nargs).invoke()

    def drop_collection(self, collection: str, nargs: Any):
        args = {
            "name": collection,
        }
        NCall(self.database_client.delete_collection, args, nargs).invoke()

    def list_collections(self, nargs) -> Any:
        return NCall(
            self.database_client.list_collections, None, nargs
        ).invoke()

    def close(self, nargs: Any) -> Any:
        pass


class OperationConverter:
    processor: ItemProcessor
    dimension: int

    def __init__(
        self,
        processor: ItemProcessor,
        dimension: int,
    ) -> None:
        self.processor = processor
        self.dimension = dimension

    def filter_metadata(self, metadata: dict | None) -> dict | None:
        if metadata is None:
            return None
        md = dict()
        for k, v in metadata.items():
            if isinstance(v, (str, int, bool)):
                md[k] = v
        return md

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        batch_get = False
        batch_put = False
        batch_delete = False
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.GET):
                batch_get = True
            elif op_parser.op_equals(StoreOperation.PUT):
                batch_put = True
            elif op_parser.op_equals(StoreOperation.DELETE):
                batch_delete = True
        if (
            (batch_get and batch_put)
            or (batch_put and batch_delete)
            or (batch_delete and batch_get)
        ):
            raise BadRequestError(
                "Batch operation supports only one operation type"
            )
        if batch_get:
            return self.convert_batch_get(op_parsers), "get"
        elif batch_put:
            return self.convert_batch_put(op_parsers), "put"
        elif batch_delete:
            return self.convert_batch_delete(op_parsers), "delete"
        raise BadRequestError("Batch operation not supported")

    def convert_batch_get(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {
            "ids": ids,
            "include": ["embeddings", "documents", "metadatas"],
        }

    def convert_batch_put(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        embeddings = []
        documents = []
        metadatas = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
            metadatas.append(self.filter_metadata(op_parser.get_metadata()))
            vector_value = VectorValue(**op_parser.get_vector_value())
            embeddings.append(vector_value.vector)
            documents.append(vector_value.content)
        return {
            "ids": ids,
            "embeddings": embeddings,
            "metadatas": metadatas,
            "documents": documents,
        }

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {
            "ids": ids,
        }

    def convert_get(self, id: str) -> dict:
        return {
            "ids": [id],
            "include": ["embeddings", "documents", "metadatas"],
        }

    def convert_put(self, id: str, value: dict, metadata: dict | None) -> dict:
        vector_value = VectorValue(**value)
        args = {
            "ids": [id],
            "embeddings": [vector_value.vector],
        }
        if vector_value.content is not None:
            args["documents"] = [vector_value.content]
        if metadata is not None:
            args["metadatas"] = [self.filter_metadata(metadata)]
        return args

    def convert_update_value(self, id: str, value: dict) -> dict:
        args: dict = {"ids": [id]}
        vector_value = VectorValue(**value)
        args["embeddings"] = [vector_value.vector]
        if vector_value.content is not None:
            args["documents"] = [vector_value.content]
        return args

    def convert_update_metadata(self, id: str, metadata: dict | None) -> dict:
        args = {"ids": [id], "metadatas": [self.filter_metadata(metadata)]}
        return args

    def convert_delete(self, id: str | None, where: Expression) -> dict:
        if id is None and where is None:
            raise BadRequestError("Id or where should be provided for delete")
        if id is not None and where is None:
            return {"ids": [id]}
        if where is not None and id is None:
            return {"where": self.convert_expr(where)}
        return {"ids": [id], "where": self.convert_expr(where)}

    def convert_query(
        self,
        search: Function | None,
        where: Expression,
        select: Select | None,
        order_by: OrderBy | None,
        limit: int | None,
        offset: int | None,
    ) -> dict:
        args: dict = dict()
        search_args = self.convert_search(search)
        args = args | search_args
        select_args = self.convert_select(select)
        if select_args is not None:
            args = args | select_args
        if where is not None:
            args["where"] = self.convert_expr(where)
        _offset = offset if offset is not None else 0
        _limit = limit if limit is not None else DEFAULT_LIMIT
        args["n_results"] = _limit + _offset
        return args

    def convert_select(self, select: Select | None) -> dict:
        include = ["distances"]
        if select is None or len(select.terms) == 0:
            include.extend(["embeddings", "documents", "metadatas"])
        else:
            for term in select.terms:
                if is_value_field(term.field):
                    include.extend(["embeddings", "documents"])
                elif is_metadata_field(term.field):
                    include.extend(["metadatas"])
        return {"include": include}

    def convert_search(self, search: Function | None) -> Any:
        if search is None:
            return {"query_embeddings": [0.0] * self.dimension}
        namespace = search.namespace
        name = search.name
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.VECTOR_SEARCH:
                args = VectorSearchArgs(**search.named_args)
                return {
                    "query_embeddings": [args.vector],
                }

        raise BadRequestError("Search function not supported")

    def convert_expr(self, expr: Expression) -> Any:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            return expr
        if isinstance(expr, Field):
            field = self.processor.resolve_metadata_field(expr.path)
            return field
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            comparion_map = {
                ComparisonOp.EQ: "$eq",
                ComparisonOp.NEQ: "$ne",
                ComparisonOp.GT: "$gt",
                ComparisonOp.GTE: "$gte",
                ComparisonOp.LT: "$lt",
                ComparisonOp.LTE: "$lte",
                ComparisonOp.IN: "$in",
                ComparisonOp.NIN: "$nin",
            }
            op = expr.op
            if isinstance(expr.lexpr, Field):
                field = self.convert_expr(expr.lexpr)
                value = self.convert_expr(expr.rexpr)
            elif isinstance(expr.rexpr, Field):
                field = self.convert_expr(expr.rexpr)
                value = self.convert_expr(expr.lexpr)
                op = Comparison.reverse_op(op)

            if op in comparion_map:
                pc_op = comparion_map[op]
                return {field: {pc_op: value}}
            elif op == ComparisonOp.BETWEEN:
                return {
                    "$and": [
                        {field: {"$gte": value[0]}},
                        {field: {"$lte": value[1]}},
                    ]
                }
            else:
                raise BadRequestError(f"Comparison op {expr.op} not supported")
        if isinstance(expr, And):
            return {
                "$and": [
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            }
        if isinstance(expr, Or):
            return {
                "$or": [
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            }
        if isinstance(expr, Not):
            raise BadRequestError("Not expression is not supported")
        return str(expr)

    def convert_func(self, expr: Function) -> dict:
        name = expr.name
        raise BadRequestError(f"Function {name} not supported")

    def convert_count(self, where: Expression) -> dict | None:
        if where is not None:
            raise BadRequestError("Where in count is not supported")
        args: dict = {}
        return args


class ChromaCollection:
    client: Any
    converter: OperationConverter
    processor: ItemProcessor

    def __init__(self, client: Any, dimension: int):
        self.client = client
        self.processor = ItemProcessor()
        self.converter = OperationConverter(
            processor=self.processor,
            dimension=dimension,
        )



"""
Secret Store on Environment variables.
"""

__all__ = ["Environment"]

import os
from typing import Any

from verse.core import Context, Operation, Response
from verse.core.exceptions import NotFoundError, PreconditionFailedError
from verse.internal.storage_core import StoreOperation, StoreProvider

from .._constants import LATEST_VERSION
from .._models import SecretItem, SecretKey, SecretVersion


class Environment(StoreProvider):
    def __init__(self, **kwargs):
        """Initalize."""
        pass

    def init(self, context: Context | None = None) -> None:
        pass

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        result = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            val = os.environ.get(id)
            if val is None:
                raise NotFoundError
            result = SecretItem(
                key=SecretKey(id=id, version=LATEST_VERSION), value=val
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            val = os.environ.get(id)
            if val is None:
                raise NotFoundError
            versions: list = []
            versions.append(SecretVersion(version=LATEST_VERSION))
            result = SecretItem(key=SecretKey(id=id), versions=versions)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = str(op_parser.get_value())
            exists = op_parser.get_where_exists()
            if exists is False:
                current_value = os.environ.get(id)
                if current_value is not None:
                    raise PreconditionFailedError
            elif exists is True:
                current_value = os.environ.get(id)
                if current_value is None:
                    raise PreconditionFailedError
            os.environ[id] = value
            result = SecretItem(key=SecretKey(id=id, version=LATEST_VERSION))
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            value = str(op_parser.get_updated_value())
            current_value = os.environ.get(id)
            if current_value is None:
                raise NotFoundError
            os.environ[id] = value
            result = SecretItem(key=SecretKey(id=id, version=LATEST_VERSION))
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            try:
                os.environ.pop(id)
            except KeyError:
                raise NotFoundError
            result = None
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            result = None
        else:
            return super().run(operation)
        return Response(result=result)
